// -*- C++ -*-
// $RCSfile: petsc_solver.C,v $
// $Revision: 1.25 $
// $Author: langer $
// $Date: 2005/12/13 16:45:21 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#ifdef HAVE_PETSC

#include <oofconfig.h>
#include "mpi.h"
#include "petscvec.h"
#include "petscmat.h"
#include "petscsles.h"
#include "petsc_solver.h"
#include "petsc_preconditioner.h"
#include "engine/csrmatrix.h"
#include "engine/matvec.h"
#include "common/argv.h"

PETScEnvironment::PETScEnvironment() : number_of_processors (0) {
  //std::cerr << "PETScEnvironment::PETScEnvironment()" << std::endl;
  char **argv = get_argv();
  
  int argc = get_argc();
  //std::cerr << "argc=" << argc << std::endl;
  /*
    for (int i = 0; i<argc; i++)
    {
    //std::cerr << " I am here! " << std::endl;
    std::cerr << "before" << argv[i] << std::endl;
    }
  */
  char help[] = "OOF2 override of PETSc";
  //int state;
  //state = MPI_Init(&argc, &argv); //initializes MPI
  //std::cerr << "state=" << state << std::endl;
  PetscInitialize(&argc, &argv,PETSC_NULL, help);//Initializes PETSc. 
  //IF MPI has not been initialized, PETSc does it.
  /*
  for (int i = 0; i<argc; i++)
    {
      //std::cerr << " I am here! " << std::endl;
      std::cerr << "AFTER" << argv[i] << std::endl;
    } 
  int size;
  MPI_Comm_size(PETSC_COMM_WORLD,&size);
  std::cerr << "number_of_processors= " << size << std::endl;
  */
}

PETScEnvironment::~PETScEnvironment() {
  PetscFinalize(); //closes any pending PETSc work
  //MPI_Finalize(); //close MPI environment
  //std::cerr << "PETScEnvironment::~PETScEnvironment()" << std::endl;
}

PETScSolver::PETScSolver() 
  : max_iterations(100000), 
  tolerance(1.0e-5),
  number_of_iterations(0),
  residual_norm(0),
  pc(0), 
  arlb(1.0e-50),
  arub(1.0e50),
  status_(0)
  //Above values for convergence are PETSc's defaults, according to the manual pg. 65. 
{

  
}


PETScSolver::PETScSolver(int maxiter, double tol, double arlb, double arub, PreconditionerWrap * pc)
  : max_iterations(maxiter), 
  tolerance(tol), 
  number_of_iterations(0),
  residual_norm(0),
  arlb(arlb),
  arub(arub),
  pc(pc)
  
{
  //std::cerr << " allocating PETScSolver::PETScSolver()" << std::endl;
 
}


PETScSolver::~PETScSolver() {
  //std::cerr << " destroying PETScSolver::~PETScSolver()" << std::endl;
  
}

void PETScSolver::precompute(const CSRmatrix &A){
  //  std::cerr << "Inside PETScSolver::precompute()" << std::endl;
}

void PETScSolver::postcompute() {
  //std::cerr << "Inside PETScSolver::postcompute()" << std::endl;
}


bool PETScSolver::solve(const CSRmatrix &A, const VECTOR_D &rhs, VECTOR_D &x)
{
  //First, transform CSRMatrix and Vector objects into Mat and Vec objects
  
  //  std::cerr << "Entering PETScSolver::solve" << std::endl;
  SLES SolverContext; /* 
			 Preconditioner and Krylov solver contexts should
			 have associated the local SolverContext.
			 In principle, we can create a SLES context
			 for each system of equation to be solved.
			 
		      */
  PC PreconditionerContext;/*
			     <--Context for associating a preconditioner
			     to the local solver.
			   */
  KSP KrylovSolverContext; /*
			     Will be used by derived classes to set
			     correct Krylov Method
			   */
  static PETScEnvironment petsc_environment; //If mesh is solved with PETSc, keep the environment until closing session
  
  
  Vec x_PETSc, rhs_PETSc;
  Mat A_PETSc;
  VecCreate(PETSC_COMM_WORLD,&x_PETSc); 
  /*
    Assume PETSC_COMM_WORLD is the ONLY way to interact with 
    MPI communicator.
    PETSC and MPI environment are set at the FEMesh level.
    Will make it better in a later pass, promise...
  */
  PetscObjectSetName((PetscObject) x_PETSc, "Solution");
  int vector_dimension = x.size();
  //std::cerr << "vector size = " << vector_dimension << std::endl;
  VecSetSizes(x_PETSc,PETSC_DECIDE,vector_dimension);
  VecSetFromOptions(x_PETSc);
  //pass rhs to rhs_PETSc
  VecDuplicate(x_PETSc,&rhs_PETSc);
  for (int i = 0; i<vector_dimension; i++)
    {
      PetscScalar temp = rhs(i);
      VecSetValues(rhs_PETSc,1, &i, &rhs(i), INSERT_VALUES);
      std::cerr << "rhs = " << temp << std::endl;
    }
  VecAssemblyBegin(rhs_PETSc);
  //anything you want to do on the mean time? Put it here.
  VecAssemblyEnd(rhs_PETSc);

  //pass A to A_PETSc
  MatCreate(PETSC_COMM_WORLD,PETSC_DECIDE,PETSC_DECIDE,vector_dimension,vector_dimension,&A_PETSc);
  MatSetFromOptions(A_PETSc);
  for (int i = 0; i<vector_dimension; i++) //This is a bottle neck and should be removed
    {
      for (int j = 0; j<vector_dimension; j++)
	{
	  double x = A(i,j);
	  
	  if (x != 0.0)
	    {
	      MatSetValues(A_PETSc, 1, &i, 1, &j, &x, INSERT_VALUES);
	      
	    }
	}
      
    }
  MatAssemblyBegin(A_PETSc, MAT_FINAL_ASSEMBLY);
  MatAssemblyEnd(A_PETSc, MAT_FINAL_ASSEMBLY);
  SLESCreate(PETSC_COMM_WORLD,&SolverContext);
  SLESSetOperators(SolverContext,A_PETSc,A_PETSc,DIFFERENT_NONZERO_PATTERN); 
  /* 
     tell PETSc that we are 
     going to solve A_PETSc matrix
  */
  //initialize Krylovmethods
  SLESGetKSP(SolverContext,&KrylovSolverContext);
  SLESGetPC(SolverContext,&PreconditionerContext); 
  //tell Solver context that we will use a preconditioner
  Preconditioner * preconditioner = pc->create_preconditioner(A, rowblocks, colblocks);
  preconditioner->set(PreconditionerContext); //call desired preconditioner  
  setsolver(KrylovSolverContext); 
  /*
    should set correct Krylov method in derived class; 
    virtual in this context. Each derived class has special requirements, which
    are accounted for by programming the right methods. 
  */
  
  KSPSetTolerances(KrylovSolverContext, tolerance, arlb, arub, max_iterations);
  /*
    set tolerance(s). PETSC_DEFAULT flags should be removed 
    and accessed by user. Will do it soon.
  */
  SLESSetFromOptions(SolverContext);
  
  
  
  SLESSolve(SolverContext,rhs_PETSc,x_PETSc,&number_of_iterations); //<--REAL work
  
  //SLESView(SolverContext,PETSC_VIEWER_STDOUT_WORLD);
  //VecView(x_PETSc,PETSC_VIEWER_STDOUT_WORLD);

  // Here, the residual and the reason for breaking-out of the convergence
  //loop are sampled
  KSPGetResidualNorm(KrylovSolverContext,&residual_norm);
  
  KSPConvergedReason reason;
  KSPGetConvergedReason(KrylovSolverContext,&reason);
  int integer_solver_state = int(reason);
  
  PETScSolverMessager messenger(integer_solver_state, number_of_iterations, residual_norm);
  if (integer_solver_state < 0)
    status_ = 1; //did NOT converge
  else
    status_ = 0;//did TOO converge
  
  _message = messenger._message();
  
  //pass local solution to global solution
  PetscScalar * temp_array;
  VecGetArray(x_PETSc, &temp_array);
  for (int i = 0; i<vector_dimension; i++)
    {
      x(i) = double(temp_array[i]);
    }
  VecRestoreArray(x_PETSc,&temp_array);


  //clear memory of temporary objects
  VecDestroy(x_PETSc);
  VecDestroy(rhs_PETSc);
  MatDestroy(A_PETSc);
  SLESDestroy(SolverContext); 
  //  std::cerr << " leaving PETScSolver::solve()" << std::endl;
  delete preconditioner;
  return !status_;
}

/*
  All PETSc implementations call the same functions, except for
  setsolver, which should be implemented on each derived class
 */

void  PETScGMRESSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPGMRES);
  //  std::cerr << "PETScGMRESSolver::setsolver()" << endl;
}

std::string PETScGMRESSolver::message() const {
  return "PETScGMRES" + _message;
}

void PETScBiCGSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPBICG);
  //std::cerr << "PETScGMRESSolver::setsolver()" << endl;
}

std::string PETScBiCGSolver::message() const {
  return "PETScBiCG" + _message;
}

void PETScRichardsonSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPRICHARDSON);
  //std::cerr << "PETScRichardsonSolver::setsolver()" << endl;
}

std::string PETScRichardsonSolver::message() const {
  return "PETScRichardson" + _message;
}


void PETScChebychevSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPCHEBYCHEV);
  //std::cerr << "PETScChebychevSolver::setsolver()" << endl;
}

std::string PETScChebychevSolver::message() const {
  return "PETScChebychev" + _message;
}

void PETScCGSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPCG);
  //std::cerr << "PETScCGSolver::setsolver()" << endl;
}

std::string PETScCGSolver::message() const {
  return "PETScCG" + _message;
}

void PETScBiCGStabSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPBCGS);
  //std::cerr << "PETScBCGStabSolver::setsolver()" << endl;
}

std::string PETScBiCGStabSolver::message() const {
  return "PETScBiCGStab" + _message;
}

void PETScCGSquaredSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPCGS);
  //std::cerr << "PETScCGSquaredSolver::setsolver()" << endl;
}

std::string PETScCGSquaredSolver::message() const {
  return "PETScCGSquared" + _message;
}

void PETScTFQMR1Solver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPTFQMR);
  //std::cerr << "PETScTFQMR1Solver::setsolver()" << endl;
}

std::string PETScTFQMR1Solver::message() const {
  return "PETScTFQMR1" + _message;
}


void PETScTFQMR2Solver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPTCQMR);
  //std::cerr << "PETScTFQMR1Solver::setsolver()" << endl;
}

std::string PETScTFQMR2Solver::message() const {
  return "PETScTFQMR2" + _message;
}


void PETScCRSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPCR);
  //std::cerr << "PETScCRSolver::setsolver()" << endl;
}

std::string PETScCRSolver::message() const {
  return "PETScCR" + _message;
}


void PETScLeastSquaresSolver::setsolver(KSP & Kcontext) const
{ 
  KSPSetType(Kcontext, KSPLSQR);
  //std::cerr << "PETScLeastSquaresSolver::setsolver()" << endl;
}

std::string PETScLeastSquaresSolver::message() const {
  return "PETScLeastSquares" + _message;
}


PETScSolverMessager::PETScSolverMessager(const int state, const int num_iter, const double res) {
  switch (state)
    {
      //These are PETSc's convergence messages
    case 0 : 
      message = " converged! \n";
      break;
    case 2:
      message = " converged! \nRelative residual smaller than specified tolerance. \n";
      break;
    case 3:
      message = " converged! \nRelative residual smaller than arlb. \n";
      break;
    case 4:
      message = " converged in ONE iteration! \n";
      break;
    case 5:
      message = " converged! \nQCG negative curve (?!). \n";
      break;
    case 6:
      message = " converged! \nQCG constrained. \n";
      break;
    case 7:
      message = " converged!\nStep length. \n";
      break;
      //These are PETSc's NOT converging messages
    case -3:
      message = " FAILED to converge. \nReached maximum number of iterations. \n";
      break;
    case -4:
      message = " FAILED to converge. \nReached specified arub. \n";
      break;
    case -5:
      message = " FAILED to converge. \nGeneric breakdown of method. Try a different solver. \n";
    case -6:
      message = " BiCG method FAILED to converge.\nInitial residual is orthogonal to preconditioned initial residual. Try a different preconditioner or a different initial guess. \n";
      break;
    case -7:
      message = " FAILED to converge. \nMesh is NOT symmetric!!! \n";
      break;
    case -8:
      message = " FAILED to converge.\nIndefinite preconditioner. Try a different one. \n";
      break;
    default:
      message = " Unknown PETSc state.\n";
      break;
     
    };
  std::string solver_info = "residual = " + to_string(res) + " number of completed iterations = " + to_string(num_iter);
  message += solver_info;
}

int create_matrix()
{
  std::cerr << "Inside PETSc function" << std::endl;
  Vec         x, b, u;      /* approx solution, RHS, exact solution */
  Mat         A;            /* linear system matrix */
  SLES        sles;         /* linear solver context */
  PC          pc;           /* preconditioner context */
  KSP         ksp;          /* Krylov subspace method context */
  PetscReal   norm;         /* norm of solution error */
  int         ierr,i,n = 10,col[3],its,rstart,rend,nlocal;
  PetscScalar neg_one = -1.0,one = 1.0,value[3];
  static PETScEnvironment petsc_environment; //If mesh is solved with PETSc, keep the environment until closing session
  std::cerr << "PETSc initialized" << std::endl;
  

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
     Compute the matrix and right-hand-side vector that define
     the linear system, Ax = b.
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  
  /* 
     Create vectors.  Note that we form 1 vector from scratch and
     then duplicate as needed. For this simple case let PETSc decide how
     many elements of the vector are stored on each processor. The second
     argument to VecSetSizes() below causes PETSc to decide.
  */
  ierr = VecCreate(PETSC_COMM_WORLD,&x);//;CHKERRQ(ierr);
  ierr = VecSetSizes(x,PETSC_DECIDE,n);//CHKERRQ(ierr);
  ierr = VecSetFromOptions(x);//CHKERRQ(ierr);
  ierr = VecDuplicate(x,&b);//CHKERRQ(ierr);
  ierr = VecDuplicate(x,&u);//CHKERRQ(ierr);
  std::cerr << "PETSc vector created " << std::endl;
  /* Identify the starting and ending mesh points on each
     processor for the interior part of the mesh. We let PETSc decide
     above. */

  ierr = VecGetOwnershipRange(x,&rstart,&rend);//CHKERRQ(ierr);
  ierr = VecGetLocalSize(x,&nlocal);//CHKERRQ(ierr);

  /* 
     Create matrix.  When using MatCreate(), the matrix format can
     be specified at runtime.

     Performance tuning note:  For problems of substantial size,
     preallocation of matrix memory is crucial for attaining good 
     performance.  Since preallocation is not possible via the generic
     matrix creation routine MatCreate(), we recommend for practical 
     problems instead to use the creation routine for a particular matrix
     format, e.g.,
         MatCreateMPIAIJ() - sequential AIJ (compressed sparse row)
         MatCreateMPIBAIJ() - block AIJ
     See the matrix chapter of the users manual for details.

     We pass in nlocal as the "local" size of the matrix to force it
     to have the same parallel layout as the vector created above.
  */
  ierr = MatCreate(PETSC_COMM_WORLD,nlocal,nlocal,n,n,&A);//CHKERRQ(ierr);
  std::cerr << "PETSc Matrix created" << std::endl;
  ierr = MatSetFromOptions(A);//CHKERRQ(ierr);
  std::cerr << "MatSetFromOptions done" << std::endl;
  /* 
     Assemble matrix.  

     The linear system is distributed across the processors by 
     chunks of contiguous rows, which correspond to contiguous
     sections of the mesh on which the problem is discretized.  
     For matrix assembly, each processor contributes entries for
     the part that it owns locally.
  */


  if (!rstart) {
    rstart = 1;
    i = 0; col[0] = 0; col[1] = 1; value[0] = 2.0; value[1] = -1.0;
    ierr = MatSetValues(A,1,&i,2,col,value,INSERT_VALUES);//CHKERRQ(ierr);
  }
  if (rend == n) {
    rend = n-1;
    i = n-1; col[0] = n-2; col[1] = n-1; value[0] = -1.0; value[1] = 2.0; 
    ierr = MatSetValues(A,1,&i,2,col,value,INSERT_VALUES);//CHKERRQ(ierr);
  }

  /* Set entries corresponding to the mesh interior */
  value[0] = -1.0; value[1] = 2.0; value[2] = -1.0;
  for (i=rstart; i<rend; i++) {
    col[0] = i-1; col[1] = i; col[2] = i+1;
    ierr = MatSetValues(A,1,&i,3,col,value,INSERT_VALUES);//CHKERRQ(ierr);
  }

  /* Assemble the matrix */
  ierr = MatAssemblyBegin(A,MAT_FINAL_ASSEMBLY);//CHKERRQ(ierr);
  ierr = MatAssemblyEnd(A,MAT_FINAL_ASSEMBLY); //CHKERRQ(ierr);

  /* 
     Set exact solution; then compute right-hand-side vector.
  */
  ierr = VecSet(&one,u); //CHKERRQ(ierr);
  ierr = MatMult(A,u,b);//CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                Create the linear solver and set various options
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* 
     Create linear solver context
  */
  ierr = SLESCreate(PETSC_COMM_WORLD,&sles);//CHKERRQ(ierr);

  /* 
     Set operators. Here the matrix that defines the linear system
     also serves as the preconditioning matrix.
  */
  ierr = SLESSetOperators(sles,A,A,DIFFERENT_NONZERO_PATTERN);//CHKERRQ(ierr);

  /* 
     Set linear solver defaults for this problem (optional).
     - By extracting the KSP and PC contexts from the SLES context,
       we can then directly call any KSP and PC routines to set
       various options.
     - The following four statements are optional; all of these
       parameters could alternatively be specified at runtime via
       SLESSetFromOptions();
  */
  ierr = SLESGetKSP(sles,&ksp);//CHKERRQ(ierr);
  ierr = SLESGetPC(sles,&pc);//CHKERRQ(ierr);
  ierr = PCSetType(pc,PCJACOBI);//CHKERRQ(ierr);
  ierr = KSPSetTolerances(ksp,1.e-7,PETSC_DEFAULT,PETSC_DEFAULT,PETSC_DEFAULT);//CHKERRQ(ierr);

  /* 
    Set runtime options, e.g.,
        -ksp_type <type> -pc_type <type> -ksp_monitor -ksp_rtol <rtol>
    These options will override those specified above as long as
    SLESSetFromOptions() is called _after_ any other customization
    routines.
  */
  ierr = SLESSetFromOptions(sles); //CHKERRQ(ierr);
 
  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                      Solve the linear system
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* 
     Solve linear system
  */
  ierr = SLESSolve(sles,b,x,&its);//CHKERRQ(ierr); 

  /* 
     View solver info; we could instead use the option -sles_view to
     print this info to the screen at the conclusion of SLESSolve().
  */
  //  ierr = SLESView(sles,PETSC_VIEWER_STDOUT_WORLD);CHKERRQ(ierr);

  /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
                      Check solution and clean up
     - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - */
  /* 
     Check the error
  */
  ierr = VecAXPY(&neg_one,u,x);//CHKERRQ(ierr);
  ierr  = VecNorm(x,NORM_2,&norm);//CHKERRQ(ierr);
  ierr = PetscPrintf(PETSC_COMM_WORLD,"Norm of error %A, Iterations %d\n",norm,its);//CHKERRQ(ierr);
  /* 
     Free work space.  All PETSc objects should be destroyed when they
     are no longer needed.
  */
  ierr = VecDestroy(x);
  ierr = VecDestroy(u);
  ierr = VecDestroy(b);//CHKERRQ(ierr); 
  ierr = MatDestroy(A);//CHKERRQ(ierr);
  ierr = SLESDestroy(sles);//CHKERRQ(ierr);

  /*
     Always call PetscFinalize() before exiting a program.  This routine
       - finalizes the PETSc libraries as well as MPI
       - provides summary and diagnostic information if certain runtime
         options are chosen (e.g., -log_summary).
  */
  //ierr = PetscFinalize();CHKERRQ(ierr);
  return 0;
}

#endif //HAVE_PETSC


