# -*- python -*-
# $RCSfile: conjugate.py,v $
# $Revision: 1.11 $
# $Author: reida $
# $Date: 2006/01/23 19:44:30 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov.

from oof2.SWIG.common import ooferror
from oof2.SWIG.common import lock
from oof2.SWIG.common import switchboard
from oof2.SWIG.engine import material
from oof2.SWIG.engine import fieldindex
from oof2.common import debug
from oof2.engine import propertyregistration
import oof2.engine.mesh

FieldIndexPtr = fieldindex.FieldIndexPtr

# TODO: There's currently no way to destroy or remove a
# ConjugatePairObj from the big master list, effectively preventing
# editing.

class ConjugatePairObj:
    def __init__(self, name, eqn, eqncomp, field, fieldcomp):
        self.name = name
        self.equation = eqn
        self.equation_comp = eqncomp
        self.field = field
        self.field_comp = fieldcomp
    def get_field(self):
        return self.field
    def get_field_component(self):
        return self.field_comp
    def get_equation(self):
        return self.equation
    def get_equation_component(self):
        return self.equation_comp
    def get_name(self):
        return self.name
    # This equality operator prevents duplication of conjugate pair
    # objects in the list.
    def __eq__(self, other):
        if other:
            if isinstance(other,ConjugatePairObj):
                if self.name==other.name and \
                       self.equation==other.equation and \
                       self.equation_comp == other.equation_comp and \
                       self.field==other.field and \
                       self.field_comp==other.field_comp:
                    return True
        return False
    def __hash__(self):
        return hash(self.name) ^ hash(self.equation) ^ \
               hash(self.equation_comp) ^ \
               hash(self.field) ^ hash(self.field_comp)
                
    

class ListOfConjugatePairs:
    def __init__(self):
        self.pairs = []
        self.number_of_pairs = 0
        self.lock = lock.Lock()

    def add(self, conjpair):
        self.lock.acquire()
        try:
            if conjpair not in self.pairs:
                self.pairs.append(conjpair)
                self.number_of_pairs += 1
        finally:
            self.lock.release()

    # There is no removal capability for this list.  The reason is
    # that there's no point -- there does not exist a scenario where
    # removing some conjugacy data helps you in any way.  If you're
    # using a bunch of exciting and strange new properties, then you
    # aren't using the regular ones, and the associated conjugacy data
    # will not be relevant, so there's no point in removing it.  If
    # you are using properties of the given type, then the conjugacy
    # data is correct.  If you have additional conjugacy data which
    # conflicts with it, this indicates a real inability to symmetrize
    # the problem, so the inconsistencies in the list reflect a real
    # physical inconsistency in the set-up of the problem.  This may
    # or may not be an error, but in any case, it is correct for this
    # list to report it.

    
    def relevant_property(self, property, pair):
        return property.is_property_type(pair.get_name())
    
    def sanity_check(self, relevant_pairs):
        # For a given list of properties, make sure that no field maps
        # to more than one equation, and vice versa.  Does not check
        # for completeness, some equations or fields may be unmapped
        # and the sanity_check will still pass.

        self.lock.acquire()
        try:
            # Loop over pairs of pairs.  If there are conflicts, in
            # that a field is conjugate to multiple equations or vice
            # versa, then the sanity check fails. 
            for i in range(len(relevant_pairs)):
                cpair1 = relevant_pairs[i]
                for j in range(i + 1, len(relevant_pairs)):
                    cpair2 = relevant_pairs[j]
                    # If equations match but fields differ, we are not sane.
                    if cpair1.get_equation() == cpair2.get_equation() \
                           and cpair1.get_equation_component() == \
                           cpair2.get_equation_component():
                        if cpair1.get_field() != cpair2.get_field() \
                               or cpair1.get_field_component() != \
                               cpair2.get_field_component():
                            return False
                    # Likewise, if fields match but equations differ,
                    # we are equally insane.
                    if cpair1.get_field()==cpair2.get_field() \
                       and cpair1.get_field_component() == \
                       cpair2.get_field_component():
                        if cpair1.get_equation() != cpair2.get_equation() \
                           or cpair1.get_equation_component() != \
                           cpair2.get_equation_component():
                            return False
                        
        finally:
            self.lock.release()
        return True

    def find_equation_mapping(self, femesh):
        ## femesh is a real FEMesh
        self.lock.acquire()
        try:
            local_row = []
            local_col = []

            # Indices here are the indices into the FuncNode's lists of
            # DOFs and equations.
            for conjugatepair in self.pairs:
                field = conjugatepair.get_field()
                fieldcomp = conjugatepair.get_field_component()
                equation = conjugatepair.get_equation()
                equationcomp = conjugatepair.get_equation_component()
                if femesh.is_active_field(field) and \
                       femesh.is_active_equation(equation):
                    eqcomp = femesh.active_equation_index(equation) + \
                             equationcomp.integer()
                    fcomp = femesh.defined_index(field) + fieldcomp.integer()
                    local_row.append(eqcomp)
                    local_col.append(fcomp)

            femesh.set_equation_mapping(local_row, local_col)
        finally:
            self.lock.release()

listofconjugatepairs = ListOfConjugatePairs()

######################################

def conjugatePair(name, equation, eqncomp, field, fieldcomp):
    # eqncomp and fieldcomp can either be a single FieldIndex object,
    # or a list of them.  The lists must have the same length.
    if isinstance(eqncomp, FieldIndexPtr):
        eqncomp = [eqncomp]
    if isinstance(fieldcomp, FieldIndexPtr):
        fieldcomp = [fieldcomp]
    if len(eqncomp) != len(fieldcomp):
        raise ooferror.ErrPyProgrammingError(
            "Bad index specification in conjugatePair")
    for (ecomp, fcomp) in zip(eqncomp, fieldcomp):
        listofconjugatepairs.add(
            ConjugatePairObj(name, equation, ecomp, field, fcomp))

######################################


def check_symmetry(meshName, *args):
    # Switchboard callback for "equation [de]activated" and "field
    # [de]activated".  Also called from the MaterialMSPlugIn, so it
    # gets run when the properties of a material in a mesh are
    # changed.  Also an indirect callback for "materials changed in
    # microstructure", see below.  meshName can be a colon-separated
    # string, or a list of strings.


    meshWho = oof2.engine.mesh.meshes[meshName]
    mesh = meshWho.getObject()
    ms = meshWho.getMicrostructure()
    
    materials = material.getMaterials(ms)
    relevant_pairs = {}


    for matter in materials:
        for property in matter.properties():
            # If any property is not symmetric on this mesh, then we
            # are unsymmetric.  Set symmetry state to unsymmetric and
            # exit.
            if not property.is_symmetric(mesh):
                meshWho.set_symmetry_state(
                    oof2.engine.mesh.SYM_STATE_UNSYMMETRIC)
                return

            # Find the conjugate pairs which are relevant to this
            # property, and add them to the dictionary of relevant
            # pairs.  A dictionary is used to guarantee uniqueness.
            p_reg = propertyregistration.AllProperties[property.name()]
                                  
            for pair in listofconjugatepairs.pairs:
                if listofconjugatepairs.relevant_property(p_reg, pair):
                    relevant_pairs[pair] = pair

    # The reason for checking equations rather than fields here is
    # that equations are the things that actually get remapped in the
    # mesh, so the process is more sensitive to equation-counting
    # errors.  Since a symmetric matrix must have fields corresponding
    # to equations, checking fields should be equivalent.  There is,
    # however, no reason to do both.
    meqlist = mesh.all_equations()
    for eq in meqlist:
        # If you make it through this loop for any equation, it means
        # that there is no conjugacy data for this equation.  Matrix
        # symmetry state is thus unknown, but you can still do the
        # mapping.
        for p in relevant_pairs.keys():
            if p.get_equation()==eq:
                break
        else:
            meshWho.set_symmetry_state(oof2.engine.mesh.SYM_STATE_UNSYMMETRIC)
            return

    # If there are no equations, or no conjugate pairs, then our
    # symmetry state is, by fiat, inconsistent.  Sophomoric arguments
    # about the symmetry of a size-zero matrix, or trees falling in
    # forests, are not admissible here.
    if len(meqlist)==0 or len(relevant_pairs)==0:
        meshWho.set_symmetry_state(oof2.engine.mesh.SYM_STATE_INCONSISTENT)
        return 
    
    # If we've gotten to here, then there are no asymmetric
    # properties, and the equation list is complete.  If the conjugate
    # pair set is consistent, we're symmetric, otherwise the state is
    # inconsistent, and the equation mapping must not be attempted.
    if listofconjugatepairs.sanity_check(relevant_pairs.keys()):
        meshWho.set_symmetry_state(oof2.engine.mesh.SYM_STATE_SYMMETRIC)
    else:
        meshWho.set_symmetry_state(oof2.engine.mesh.SYM_STATE_INCONSISTENT)


# Direct callback for the "materials changed in microstructure" signal.
def check_wrapper(ms):
    ms_name = ms.name()
    meshpaths = oof2.engine.mesh.meshes.keys(base=ms_name)
    for meshpath in meshpaths:
        check_symmetry([ms_name]+meshpath)

switchboard.requestCallback("materials changed in microstructure",
                            check_wrapper)

    
switchboard.requestCallback("equation activated", check_symmetry)
switchboard.requestCallback("equation deactivated", check_symmetry)

switchboard.requestCallback("field activated", check_symmetry)
switchboard.requestCallback("field deactivated", check_symmetry)

