# -*- python -*-
# $RCSfile: problemmanager.py,v $
# $Revision: 1.12 $
# $Author: redwing $
# $Date: 2005/07/01 12:52:29 $

# This software was produced by NIST, an agency of the U.S. government,
# and by statute is not subject to copyright in the United States.
# Recipients of this software assume all responsibilities associated
# with its operation, modification and maintenance. However, to
# facilitate maintenance we ask that before distributing modified
# versions of this software, you first contact the authors at
# oof_manager@ctcms.nist.gov. 

from oof2.SWIG.common import switchboard
from oof2.SWIG.engine import subproblem
from oof2.common import registeredclass
from oof2.common.IO import whoville
import oof2.engine.mesh

## A ProblemManager object (and associated instances of subclasses)
## holds a list of subproblems that will be solved for.
## A global solution is obtained by calculating a tentative solution
## in each of them, which is accepted/rejected by the TimeStepAdapterManager
##
## Subclasses know how to solve the associated list of subproblems
## Base class only knows how to initialize a self.iterativeList

class ProblemManager (registeredclass.RegisteredClass):
    registry = []
    ##
    ## base class only knows how to initialize and groom itself.
    ##
    ## subclasses know how to divide every day life problems into subproblems and solve them
    ##
    def __init__(self, name, mesh):
        self.name = name
        self.mesh = mesh ## actually a mesh name
        self.iterativeList = [] ## calculations that are iterated over
        self.originalsettings = None ## actually, a subproblem object
    def set_iterativeList(self, ilist):
        self.iterativeList = ilist
    def add_item(self, item):
        self.iterativeList.append(item)
    def initialize(self):
        ## creates a subproblem object with the settings of the original
        ## mesh, if it does not exist
        meshcontext = oof2.engine.mesh.meshes[self.mesh]
        femesh =  meshcontext.femesh()
        if self.originalsettings is None:
            self.originalsettings = subproblem.SubProblem("__original_settings__")
            ## first loop over equations
            for eqn in femesh.all_equations():
                if femesh.is_active_equation(eqn):
                    ## save settings
                    self.originalsettings.addEquation(eqn)
                if femesh.is_kinetically_active_equation(eqn):
                        ## save settings
                        self.originalsettings.activateKinetics(eqn)
                if femesh.is_dynamically_active_equation(eqn):
                        ## save settings
                        self.originalsettings.activateDynamics(eqn)
            ## next loop over defined AND active fields
            for fld in femesh.all_fields():
                if femesh.is_defined_field(fld) and femesh.is_active_field(fld):
                    ## save settings
                    self.originalsettings.addField(fld)
        ## then deactivate such settings so subproblem can apply their own
        for eqn in femesh.all_equations():
            if femesh.is_active_equation(eqn):
                femesh.deactivate_equation(eqn)
                ## tell the world
                switchboard.notify('equation activated', self.mesh, eqn.name(), 0)
                if femesh.is_kinetically_active_equation(eqn):
                    femesh.kinetic_deactivate_equation(eqn)
                    ## tell the world
                    switchboard.notify('kinetics activated', self.mesh, eqn.name(), 0)
                if femesh.is_dynamically_active_equation(eqn):
                    ## deactivate
                    femesh.deactivate_dynamics(eqn)
                    ## tell the world
                    switchboard.notify('dynamics activated', self.mesh, eqn.name(), 0)
        ## next loop over defined AND active fields
        for fld in femesh.all_fields():
            if femesh.is_defined_field(fld) and femesh.is_active_field(fld):
                femesh.deactivate_field(fld)
                ##tell the world
                switchboard.notify("field activated", self.mesh, fld.name(), 0)
    def apply(self, timedrivermanager):
        pass ## BASE CLASS DOES NOT KNOW HOW TO SOLVE A PROBLEM
    
    def finalize(self, smesh):
        ## original state of global mesh is restored
        if self.originalsettings is None:
            raise Exception("%$#@! Original state of mesh has not been set!!!")
        ## first restore equations
        meshcontext = oof2.engine.mesh.meshes[self.mesh]
        femesh =  meshcontext.femesh()
        
        for eqn in femesh.all_equations():
            ## first deactivate settings of subproblem
            if eqn.name() in smesh.active_equations:
                self.mesh.deactivate_equation(eqn)
                switchboard.notify('equation activated', self.mesh, eqn.name(), 0)
            if eqn.name() in smesh.active_kinetics:
                femesh.deactivate_kinetics(eqn)
                switchboard.notify('kinetics activated', self.mesh, eqn.name(), 0)
            if eqn.name() in smesh.active_dynamics:
                femesh.deactivate_dynamics(eqn)
                switchboard.notify('dynamics activated', self.mesh, eqn.name(), 0)
            ## next put back settings of original mesh
            if eqn.name() in self.originalsettings.active_equations:
                femesh.activate_equation(eqn)
                switchboard.notify('equation activated', self.mesh, eqn.name(), 1)
                if eqn.name() in self.originalsettings.active_kinetics:
                    femesh.activate_kinetics(eqn)
                    switchboard.notify('kinetics activated', self.mesh, eqn.name(), 1)
                if eqn.name() in self.originalsettings.active_dynamics:
                    femesh.activate_dynamics(eqn)
                    switchboard.notify('dynamics activated', self.mesh, eqn.name(), 1)
        ## next restore active fields:
        for fld in femesh.all_fields():
            if femesh.is_defined_field(fld) and femesh.is_active_field(fld):
                femesh.deactivate_field(fld)
                switchboard.notify("field activated", self.mesh, fld.name(), 0)
            if fld.name() in self.originalsettings.active_fields:
                femesh.activate_field(fld)
                switchboard.notify("field activated", self.mesh, fld.name(), 1)
        ## elements and associated DOFs and nodal equations
        ## are set to be unselected (unsubmeshable)
        skelname, group = smesh.getActiveRegion()
        skelc = whoville.getClass('Skeleton')[skelname]
        members = skelc.elementgroups.get_group(group)
        for skel_el in members:
            element = self.mesh.getElement(skel_el.meshindex)
            element.unsetsubmesh()
    def cleanup(self):
        meshcontext = oof2.engine.mesh.meshes[self.mesh]
        meshcontext.solved()
        switchboard.notify('redraw')
        ## delete settings last
        del self.originalsettings


class SubProblemIterator (ProblemManager):
    def __init__(self, name, mesh):
        ProblemManager.__init__(self, name, mesh)
    def iterationLoop(self, timeadaptermanager):
        while timeadaptermanager.iterate(self.mesh, self.iterativeList):
            ## tdriver should take a decision on keep iterating or not
            ## based (or not) on the individual time drivers in each
            ## subproblem
            
            ## individual time drivers are applied
            ## solutoin is NOT accepted
            timeadaptermanager.apply(self.initialize, self.finalize, self.mesh, self.iterativeList)

            ## the state of the set if tentative solutions is evaluated
            ## time is increased here, if solutions are accepted
            timeadaptermanager.update(self.initialize, self.finalize, self.mesh, self.iterativeList)
    
    def apply(self, timeadaptermanager):
        self.iterationLoop(timeadaptermanager)
        self.cleanup() ## should update solution time stamp and redraw
        
registeredclass.Registration(
    'SubProblemIterator',
    ProblemManager,
    SubProblemIterator,
    ordering = 100,
    params = [
    parameter.StringParameter('name',"problem_name"),
    whoville.WhoParameter('mesh', oof2.engine.mesh.meshes,
                          tip=parameter.emptyTipString)
    ],
    tip = 'applies selected solvers to a list of regions in a mesh'
    )
