// -*- C++ -*-
// $RCSfile: stiffnessmatrix.C,v $
// $Revision: 1.16 $
// $Author: haan $
// $Date: 2005/11/15 23:14:58 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

// The StiffnessMatrix class stores the components of the global
// stiffness matrix.  The components are accessed by equation number
// (as returned by Node::eqno() and by degree of freedom index (as
// returned by DegreeOfFreedom::dofindex()).  The StiffnessMatrix can
// compute submatrices corresponding to the fixed and free equations
// and degrees of freedom.  It returns these as compressed sparse row
// (CSRmatrix) matrices.

#include <oofconfig.h>
#include <iostream>
#include <vector>
#include "csrmatrix.h"
#include "femesh.h"
#include "sparselink.h"
#include "stiffnessmatrix.h"
#include "common/tostring.h"
#include "common/trace.h"

StiffnessMatrix::StiffnessMatrix(FEMesh *mush)
  : SparseLinkMatrix<double>(),
    mesh(mush)
{}

StiffnessMatrix::StiffnessMatrix(FEMesh *mush, int nr, int nc)
  : SparseLinkMatrix<double>(nr, nc),
    mesh(mush), rowgroup(nr), colgroup(nc)
{
  Trace("StiffnessMatrix::StiffnessMatrix "
 	+ to_string(nr) + "x" + to_string(nc));
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

CSRmatrix StiffnessMatrix::submatrix(const std::vector<int> &rowmap,
 				     const std::vector<int> &colmap) const
{
//   int rmax = 0, cmax=0;
//   for(int i=0; i<rowmap.size(); i++)
//     if(rowmap[i] > rmax) rmax = rowmap[i];
//   for(int i=0; i<colmap.size(); i++)
//     if(colmap[i] > cmax) cmax = colmap[i];
//   std::cerr << "StiffnessMatrix::submatrix: " << this << " " << nrows() << "x" << ncols()
// 	    << " rmax=" << rmax << " cmax=" << cmax << std::endl;
  return CSRmatrix(*this, rowmap, colmap);
}

std::vector<int> StiffnessMatrix::subgroup(const std::vector<int> groupset,
					   const std::vector<int> &map) const 
  // The "groupset" is a list, indexed by DOF, which is -1 if the DOF
  // is in-plane, and is equal to the node number (>=0) if the DOF is
  // out of plane.  The "map" is a list, also indexed by DOF, which is
  // -1 if the DOF will not be included in the submatrix, and
  // otherwise is equal to the column number of the submatrix
  // corresponding to this DOF.  Because of floating BCs, there can be
  // duplicates in the map, but these duplicates will always be
  // in-plane.  Likewise for Equations/rows.
  //   The output of this routine should be a third list, indexed by
  // column of the submatrix, with values of -1 for in-plane DOFs,
  // and the node number for out-of-plane DOFs,
{
  int subgroup_list_size = 0;
  for(std::vector<int>::size_type i=0; i<map.size(); i++) {
    if (map[i] >= subgroup_list_size) subgroup_list_size = map[i];
  }
  subgroup_list_size++;

  // Invert the map -- in case of duplicates, retain only the DOF/eqn
  // number of the first occurrence.  This is what the submatrix
  // extraction code does.
  std::vector<int> map_inverse(subgroup_list_size,-1);
  for(std::vector<int>::size_type i=0; i<map.size(); i++) {
    if (map[i]>=0) {
      if (map_inverse[map[i]] == -1) { // First occurrence only happens here.
	map_inverse[map[i]]=i;
      }
    }
  }

  std::vector<int> res(subgroup_list_size);

  for(std::vector<int>::size_type i=0; i<res.size(); i++) {
    res[i] = groupset[map_inverse[i]];
  }

  return res;
}

// Extract the subset of the rowgroups corresponding to the given rowmap.
std::vector<int> StiffnessMatrix::rowgroups(const std::vector<int> &rowmap)
  const 
{
  return subgroup(rowgroup, rowmap);
}


// Extract the subset of colgroups corresponding to the given map.
std::vector<int> StiffnessMatrix::colgroups(const std::vector<int> &colmap)
  const 
{
  return subgroup(colgroup, colmap);
}

